# Copyright (C) 2017-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict, List, Optional, Tuple

from swh.model.swhids import CoreSWHID

from .cache import VaultCache


class InMemoryVaultBackend:
    """Stub vault backend, for use in the CLI."""

    def __init__(self):
        self._cache = VaultCache(cls="memory")

    def fetch(self, bundle_type: str, swhid: CoreSWHID) -> Optional[bytes]:
        return self._cache.get(bundle_type, swhid)

    def cook(
        self, bundle_type: str, swhid: CoreSWHID, email: Optional[str] = None
    ) -> Dict[str, Any]:
        raise NotImplementedError("InMemoryVaultBackend.cook()")

    def progress(self, bundle_type: str, swhid: CoreSWHID):
        raise NotImplementedError("InMemoryVaultBackend.progress()")

    # Cookers endpoints

    def set_progress(self, bundle_type: str, swhid: CoreSWHID, progress: str) -> None:
        pass

    def set_status(self, bundle_type: str, swhid: CoreSWHID, status: str) -> None:
        pass

    def put_bundle(self, bundle_type: str, swhid: CoreSWHID, bundle) -> bool:
        self._cache.add(bundle_type, swhid, bundle)
        return True

    def send_notif(self, bundle_type: str, swhid: CoreSWHID):
        pass

    # Batch endpoints

    def batch_cook(self, batch: List[Tuple[str, str]]) -> int:
        raise NotImplementedError("InMemoryVaultBackend.batch_cook()")

    def batch_progress(self, batch_id: int) -> Dict[str, Any]:
        pass



# Copyright (C) 2015-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import logging
from typing import TYPE_CHECKING, Optional

import click

from swh.core.cli import CONTEXT_SETTINGS, AliasedGroup
from swh.core.cli import swh as swh_cli_group

if TYPE_CHECKING:
    import io

    from swh.model.swhids import CoreSWHID


class SwhidParamType(click.ParamType):
    name = "swhid"

    def convert(self, value, param, ctx):
        from swh.model.exceptions import ValidationError
        from swh.model.swhids import CoreSWHID

        try:
            return CoreSWHID.from_string(value)
        except ValidationError:
            self.fail(f"expected core SWHID, got {value!r}", param, ctx)


@swh_cli_group.group(name="vault", context_settings=CONTEXT_SETTINGS, cls=AliasedGroup)
@click.pass_context
def vault(ctx):
    """Software Heritage Vault tools."""


@vault.command()
@click.option(
    "--config-file",
    "-C",
    default=None,
    metavar="CONFIGFILE",
    type=click.Path(
        exists=True,
        dir_okay=False,
    ),
    help="Configuration file.",
)
@click.argument("swhid", type=SwhidParamType())
@click.argument("outfile", type=click.File("wb"))
@click.option(
    "--bundle-type",
    type=click.Choice(["flat", "gitfast", "git_bare"]),
    help="Selects which cooker to use, when there is more than one available "
    "for the given object type.",
)
@click.pass_context
def cook(
    ctx,
    config_file: str,
    swhid: CoreSWHID,
    outfile: io.RawIOBase,
    bundle_type: Optional[str],
):
    """
    Runs a vault cooker for a single object (identified by a SWHID),
    and outputs it to the given file.
    """
    from swh.core import config
    from swh.model.swhids import ObjectType
    from swh.objstorage.exc import ObjNotFoundError
    from swh.objstorage.factory import get_objstorage
    from swh.storage import get_storage

    from .cookers import get_cooker_cls
    from .in_memory_backend import InMemoryVaultBackend

    conf = config.read(config_file)

    try:
        from swh.graph.client import RemoteGraphClient  # optional dependency

        graph = RemoteGraphClient(**conf["graph"]) if conf.get("graph") else None
    except ModuleNotFoundError:
        if conf.get("graph"):
            raise EnvironmentError(
                "Graph configuration required but module is not installed."
            )
        else:
            graph = None

    backend = InMemoryVaultBackend()

    if bundle_type is None:
        if swhid.object_type in (
            ObjectType.RELEASE,
            ObjectType.SNAPSHOT,
        ):
            bundle_type = "git_bare"
        elif swhid.object_type in (ObjectType.DIRECTORY,):
            bundle_type = "flat"
        else:
            raise click.ClickException(
                "No default bundle type for this kind of object, "
                "use --bundle-type to choose one"
            )

    try:
        cooker_cls = get_cooker_cls(bundle_type, swhid.object_type)
    except ValueError as e:
        raise click.ClickException(*e.args)

    storage = get_storage(**conf["storage"])
    objstorage = get_objstorage(**conf["objstorage"]) if "objstorage" in conf else None
    cooker = cooker_cls(
        swhid=swhid,
        backend=backend,
        storage=storage,
        graph=graph,
        objstorage=objstorage,
        max_bundle_size=None,  # No need for a size limit, we are running locally
    )
    cooker.cook()

    try:
        bundle = backend.fetch(cooker_cls.BUNDLE_TYPE, swhid)
    except ObjNotFoundError:
        bundle = None
    if bundle is None:
        import pdb

        pdb.set_trace()
        raise click.ClickException("Cooker did not write a bundle to the backend.")
    outfile.write(bundle)


@vault.command(name="rpc-serve")
@click.option(
    "--config-file",
    "-C",
    default=None,
    metavar="CONFIGFILE",
    type=click.Path(
        exists=True,
        dir_okay=False,
    ),
    help="Configuration file.",
)
@click.option(
    "--host",
    default="0.0.0.0",
    metavar="IP",
    show_default=True,
    help="Host ip address to bind the server on",
)
@click.option(
    "--port",
    default=5005,
    type=click.INT,
    metavar="PORT",
    help="Binding port of the server",
)
@click.option(
    "--debug/--no-debug",
    default=True,
    help="Indicates if the server should run in debug mode",
)
@click.pass_context
def serve(ctx, config_file, host, port, debug):
    """Software Heritage Vault RPC server."""
    from swh.vault.api.server import make_app_from_configfile

    ctx.ensure_object(dict)
    if "log_level" in ctx.obj:
        logging.getLogger("werkzeug").setLevel(ctx.obj["log_level"])

    try:
        app = make_app_from_configfile(config_file, debug=debug)
    except EnvironmentError as e:
        click.echo(e.msg, err=True)
        ctx.exit(1)

    app.run(host, port=int(port), debug=debug)


def main():
    logging.basicConfig()
    return serve(auto_envvar_prefix="SWH_VAULT")


if __name__ == "__main__":
    main()



# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections
from email.mime.text import MIMEText
import smtplib
from typing import Any, Dict, List, Optional, Tuple

import psycopg2.extras
import psycopg2.pool

from swh.core.db import BaseDb
from swh.core.db.common import db_transaction
from swh.model.swhids import CoreSWHID
from swh.scheduler import get_scheduler
from swh.scheduler.utils import create_oneshot_task_dict
from swh.storage import get_storage
from swh.vault.cache import VaultCache
from swh.vault.cookers import COOKER_TYPES, get_cooker_cls
from swh.vault.exc import NotFoundExc

cooking_task_name = "swh.vault.cooking_tasks.SWHCookingTask"

NOTIF_EMAIL_FROM = '"Software Heritage Vault" ' "<bot@softwareheritage.org>"
NOTIF_EMAIL_SUBJECT_SUCCESS = "Bundle ready: {bundle_type} {short_id}"
NOTIF_EMAIL_SUBJECT_FAILURE = "Bundle failed: {bundle_type} {short_id}"

NOTIF_EMAIL_BODY_SUCCESS = """
You have requested the following bundle from the Software Heritage
Vault:

Bundle Type: {bundle_type}
Object SWHID: {swhid}

This bundle is now available for download at the following address:

{url}

Please keep in mind that this link might expire at some point, in which
case you will need to request the bundle again.

--\x20
The Software Heritage Developers
"""

NOTIF_EMAIL_BODY_FAILURE = """
You have requested the following bundle from the Software Heritage
Vault:

Bundle Type: {bundle_type}
Object SWHID: {swhid}

This bundle could not be cooked for the following reason:

{progress_msg}

We apologize for the inconvenience.

--\x20
The Software Heritage Developers
"""


class VaultBackend:
    """
    Backend for the Software Heritage Vault.
    """

    current_version = 4

    def __init__(self, **config):
        self.config = config
        self.cache = VaultCache(**config["cache"])
        self.scheduler = get_scheduler(**config["scheduler"])
        self.storage = get_storage(**config["storage"])
        self.smtp_server = smtplib.SMTP(**config.get("smtp", {}))

        if "db" not in self.config:
            raise ValueError(
                "The 'db' configuration entry is missing "
                "in the vault configuration file"
            )
        db_conn = config["db"]
        self._pool = psycopg2.pool.ThreadedConnectionPool(
            config.get("min_pool_conns", 1),
            config.get("max_pool_conns", 10),
            db_conn,
            cursor_factory=psycopg2.extras.RealDictCursor,
        )
        self._db = None

    def get_db(self):
        if self._db:
            return self._db
        return BaseDb.from_pool(self._pool)

    def put_db(self, db):
        if db is not self._db:
            db.put_conn()

    @db_transaction()
    def progress(
        self,
        bundle_type: str,
        swhid: CoreSWHID,
        raise_notfound: bool = True,
        db=None,
        cur=None,
    ) -> Optional[Dict[str, Any]]:
        cur.execute(
            """
            SELECT id, type, swhid, task_id, task_status, sticky,
                   ts_created, ts_done, ts_last_access, progress_msg
            FROM vault_bundle
            WHERE type = %s AND swhid = %s""",
            (bundle_type, str(swhid)),
        )
        res = cur.fetchone()
        if not res:
            if raise_notfound:
                raise NotFoundExc(f"{bundle_type} {swhid} was not found.")
            return None
        res["swhid"] = CoreSWHID.from_string(res["swhid"])
        return res

    def _send_task(self, bundle_type: str, swhid: CoreSWHID):
        """Send a cooking task to the celery scheduler"""
        task = create_oneshot_task_dict("cook-vault-bundle", bundle_type, str(swhid))
        added_tasks = self.scheduler.create_tasks([task])
        return added_tasks[0]["id"]

    @db_transaction()
    def create_task(
        self,
        bundle_type: str,
        swhid: CoreSWHID,
        sticky: bool = False,
        db=None,
        cur=None,
    ):
        """Create and send a cooking task"""
        cooker_class = get_cooker_cls(bundle_type, swhid.object_type)
        cooker = cooker_class(swhid, backend=self, storage=self.storage)

        if not cooker.check_exists():
            raise NotFoundExc(f"{bundle_type} {swhid} was not found.")

        cur.execute(
            """
            INSERT INTO vault_bundle (type, swhid, sticky)
            VALUES (%s, %s, %s)""",
            (bundle_type, str(swhid), sticky),
        )
        db.conn.commit()

        task_id = self._send_task(bundle_type, swhid)

        cur.execute(
            """
            UPDATE vault_bundle
            SET task_id = %s
            WHERE type = %s AND swhid = %s""",
            (task_id, bundle_type, str(swhid)),
        )

    @db_transaction()
    def add_notif_email(
        self, bundle_type: str, swhid: CoreSWHID, email: str, db=None, cur=None
    ):
        """Add an e-mail address to notify when a given bundle is ready"""
        cur.execute(
            """
            INSERT INTO vault_notif_email (email, bundle_id)
            VALUES (%s, (SELECT id FROM vault_bundle
                         WHERE type = %s AND swhid = %s))""",
            (email, bundle_type, str(swhid)),
        )

    def put_bundle(self, bundle_type: str, swhid: CoreSWHID, bundle) -> bool:
        self.cache.add(bundle_type, swhid, bundle)
        return True

    @db_transaction()
    def cook(
        self,
        bundle_type: str,
        swhid: CoreSWHID,
        *,
        sticky: bool = False,
        email: Optional[str] = None,
        db=None,
        cur=None,
    ) -> Dict[str, Any]:
        info = self.progress(bundle_type, swhid, raise_notfound=False)

        if bundle_type not in COOKER_TYPES:
            raise NotFoundExc(f"{bundle_type} is an unknown type.")

        # If there's a failed bundle entry, delete it first.
        if info is not None and info["task_status"] == "failed":
            cur.execute(
                "DELETE FROM vault_bundle WHERE type = %s AND swhid = %s",
                (bundle_type, str(swhid)),
            )
            db.conn.commit()
            info = None

        # If there's no bundle entry, create the task.
        if info is None:
            self.create_task(bundle_type, swhid, sticky)

        if email is not None:
            # If the task is already done, send the email directly
            if info is not None and info["task_status"] == "done":
                self.send_notification(
                    None, email, bundle_type, swhid, info["task_status"]
                )
            # Else, add it to the notification queue
            else:
                self.add_notif_email(bundle_type, swhid, email)

        return self.progress(bundle_type, swhid)

    @db_transaction()
    def batch_cook(
        self, batch: List[Tuple[str, str]], db=None, cur=None
    ) -> Dict[str, int]:
        # Import execute_values at runtime only, because it requires
        # psycopg2 >= 2.7 (only available on postgresql servers)
        from psycopg2.extras import execute_values

        for bundle_type, _ in batch:
            if bundle_type not in COOKER_TYPES:
                raise NotFoundExc(f"{bundle_type} is an unknown type.")

        cur.execute(
            """
            INSERT INTO vault_batch (id)
            VALUES (DEFAULT)
            RETURNING id"""
        )
        batch_id = cur.fetchone()["id"]

        # Delete all failed bundles from the batch
        cur.execute(
            """
            DELETE FROM vault_bundle
            WHERE task_status = 'failed'
              AND (type, swhid) IN %s""",
            (tuple(batch),),
        )

        # Insert all the bundles, return the new ones
        execute_values(
            cur,
            """
            INSERT INTO vault_bundle (type, swhid)
            VALUES %s ON CONFLICT DO NOTHING""",
            batch,
        )

        # Get the bundle ids and task status
        cur.execute(
            """
            SELECT id, type, swhid, task_id FROM vault_bundle
            WHERE (type, swhid) IN %s""",
            (tuple(batch),),
        )
        bundles = cur.fetchall()

        # Insert the batch-bundle entries
        batch_id_bundle_ids = [(batch_id, row["id"]) for row in bundles]
        execute_values(
            cur,
            """
            INSERT INTO vault_batch_bundle (batch_id, bundle_id)
            VALUES %s ON CONFLICT DO NOTHING""",
            batch_id_bundle_ids,
        )
        db.conn.commit()

        # Get the tasks to fetch
        batch_new = [
            (row["type"], CoreSWHID.from_string(row["swhid"]))
            for row in bundles
            if row["task_id"] is None
        ]

        # Send the tasks
        args_batch = [(bundle_type, swhid) for bundle_type, swhid in batch_new]
        # TODO: change once the scheduler handles priority tasks
        tasks = [
            create_oneshot_task_dict("swh-vault-batch-cooking", *args)
            for args in args_batch
        ]

        added_tasks = self.scheduler.create_tasks(tasks)
        tasks_ids_bundle_ids = [
            (task_id, bundle_type, swhid)
            for task_id, (bundle_type, swhid) in zip(
                [task["id"] for task in added_tasks], batch_new
            )
        ]

        # Update the task ids
        execute_values(
            cur,
            """
            UPDATE vault_bundle
            SET task_id = s_task_id
            FROM (VALUES %s) AS sub (s_task_id, s_type, s_swhid)
            WHERE type = s_type::cook_type AND swhid = s_swhid """,
            tasks_ids_bundle_ids,
        )
        return {"id": batch_id}

    @db_transaction()
    def batch_progress(self, batch_id: int, db=None, cur=None) -> Dict[str, Any]:
        cur.execute(
            """
            SELECT vault_bundle.id as id,
                   type, swhid, task_id, task_status, sticky,
                   ts_created, ts_done, ts_last_access, progress_msg
            FROM vault_batch_bundle
            LEFT JOIN vault_bundle ON vault_bundle.id = bundle_id
            WHERE batch_id = %s""",
            (batch_id,),
        )
        bundles = cur.fetchall()
        if not bundles:
            raise NotFoundExc(f"Batch {batch_id} does not exist.")

        for bundle in bundles:
            bundle["swhid"] = CoreSWHID.from_string(bundle["swhid"])

        counter = collections.Counter(b["status"] for b in bundles)
        res = {
            "bundles": bundles,
            "total": len(bundles),
            **{k: 0 for k in ("new", "pending", "done", "failed")},
            **dict(counter),
        }

        return res

    @db_transaction()
    def is_available(self, bundle_type: str, swhid: CoreSWHID, db=None, cur=None):
        """Check whether a bundle is available for retrieval"""
        info = self.progress(bundle_type, swhid, raise_notfound=False, cur=cur)
        return (
            info is not None
            and info["task_status"] == "done"
            and self.cache.is_cached(bundle_type, swhid)
        )

    @db_transaction()
    def fetch(
        self, bundle_type: str, swhid: CoreSWHID, raise_notfound=True, db=None, cur=None
    ) -> Optional[bytes]:
        """Retrieve a bundle from the cache"""
        available = self.is_available(bundle_type, swhid, cur=cur)
        if not available:
            if raise_notfound:
                raise NotFoundExc(f"{bundle_type} {swhid} is not available.")
            return None
        self.update_access_ts(bundle_type, swhid, cur=cur)
        return self.cache.get(bundle_type, swhid)

    @db_transaction()
    def update_access_ts(self, bundle_type: str, swhid: CoreSWHID, db=None, cur=None):
        """Update the last access timestamp of a bundle"""
        cur.execute(
            """
            UPDATE vault_bundle
            SET ts_last_access = NOW()
            WHERE type = %s AND swhid = %s""",
            (bundle_type, str(swhid)),
        )

    @db_transaction()
    def set_status(
        self, bundle_type: str, swhid: CoreSWHID, status: str, db=None, cur=None
    ) -> bool:
        req = (
            """
               UPDATE vault_bundle
               SET task_status = %s """
            + (""", ts_done = NOW() """ if status == "done" else "")
            + """WHERE type = %s AND swhid = %s"""
        )
        cur.execute(req, (status, bundle_type, str(swhid)))
        return True

    @db_transaction()
    def set_progress(
        self, bundle_type: str, swhid: CoreSWHID, progress: str, db=None, cur=None
    ) -> bool:
        cur.execute(
            """
            UPDATE vault_bundle
            SET progress_msg = %s
            WHERE type = %s AND swhid = %s""",
            (progress, bundle_type, str(swhid)),
        )
        return True

    @db_transaction()
    def send_notif(self, bundle_type: str, swhid: CoreSWHID, db=None, cur=None) -> bool:
        cur.execute(
            """
            SELECT vault_notif_email.id AS id, email, task_status, progress_msg
            FROM vault_notif_email
            INNER JOIN vault_bundle ON bundle_id = vault_bundle.id
            WHERE vault_bundle.type = %s AND vault_bundle.swhid = %s""",
            (bundle_type, str(swhid)),
        )
        for d in cur:
            self.send_notification(
                d["id"],
                d["email"],
                bundle_type,
                swhid,
                status=d["task_status"],
                progress_msg=d["progress_msg"],
            )
        return True

    @db_transaction()
    def send_notification(
        self,
        n_id: Optional[int],
        email: str,
        bundle_type: str,
        swhid: CoreSWHID,
        status: str,
        progress_msg: Optional[str] = None,
        db=None,
        cur=None,
    ) -> None:
        """Send the notification of a bundle to a specific e-mail"""
        short_id = swhid.object_id.hex()[:7]

        # TODO: instead of hardcoding this, we should probably:
        # * add a "fetch_url" field in the vault_notif_email table
        # * generate the url with flask.url_for() on the web-ui side
        # * send this url as part of the cook request and store it in
        #   the table
        # * use this url for the notification e-mail
        url = "https://archive.softwareheritage.org/api/1/vault/{}/{}/" "raw".format(
            bundle_type, swhid
        )

        if status == "done":
            text = NOTIF_EMAIL_BODY_SUCCESS.strip()
            text = text.format(bundle_type=bundle_type, swhid=swhid, url=url)
            msg = MIMEText(text)
            msg["Subject"] = NOTIF_EMAIL_SUBJECT_SUCCESS.format(
                bundle_type=bundle_type, short_id=short_id
            )
        elif status == "failed":
            text = NOTIF_EMAIL_BODY_FAILURE.strip()
            text = text.format(
                bundle_type=bundle_type, swhid=swhid, progress_msg=progress_msg
            )
            msg = MIMEText(text)
            msg["Subject"] = NOTIF_EMAIL_SUBJECT_FAILURE.format(
                bundle_type=bundle_type, short_id=short_id
            )
        else:
            raise RuntimeError(
                "send_notification called on a '{}' bundle".format(status)
            )

        msg["From"] = NOTIF_EMAIL_FROM
        msg["To"] = email

        self._smtp_send(msg)

        if n_id is not None:
            cur.execute(
                """
                DELETE FROM vault_notif_email
                WHERE id = %s""",
                (n_id,),
            )

    def _smtp_send(self, msg: MIMEText):
        # Reconnect if needed
        try:
            status = self.smtp_server.noop()[0]
        except smtplib.SMTPException:
            status = -1
        if status != 250:
            self.smtp_server.connect("localhost", 25)

        # Send the message
        self.smtp_server.send_message(msg)

    @db_transaction()
    def _cache_expire(self, cond, *args, db=None, cur=None) -> None:
        """Low-level expiration method, used by cache_expire_* methods"""
        # Embedded SELECT query to be able to use ORDER BY and LIMIT
        cur.execute(
            """
            DELETE FROM vault_bundle
            WHERE ctid IN (
                SELECT ctid
                FROM vault_bundle
                WHERE sticky = false
                {}
            )
            RETURNING type, swhid
            """.format(
                cond
            ),
            args,
        )

        for d in cur:
            self.cache.delete(d["type"], CoreSWHID.from_string(d["swhid"]))

    @db_transaction()
    def cache_expire_oldest(self, n=1, by="last_access", db=None, cur=None) -> None:
        """Expire the `n` oldest bundles"""
        assert by in ("created", "done", "last_access")
        filter = """ORDER BY ts_{} LIMIT {}""".format(by, n)
        return self._cache_expire(filter)

    @db_transaction()
    def cache_expire_until(self, date, by="last_access", db=None, cur=None) -> None:
        """Expire all the bundles until a certain date"""
        assert by in ("created", "done", "last_access")
        filter = """AND ts_{} <= %s""".format(by)
        return self._cache_expire(filter, date)



# Copyright (C) 2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import tempfile
from unittest.mock import MagicMock

import click
import click.testing
import pytest

from swh.core.cli.db import db as swhdb
from swh.core.db import BaseDb
from swh.core.db.db_utils import swh_db_module, swh_db_version
from swh.core.db.tests.test_cli import craft_conninfo
from swh.model.swhids import CoreSWHID
from swh.vault.backend import VaultBackend
from swh.vault.cli import vault as vault_cli_group
from swh.vault.cookers.base import BaseVaultCooker
from swh.vault.in_memory_backend import InMemoryVaultBackend


def test_cook_unsupported_swhid():
    runner = click.testing.CliRunner()

    result = runner.invoke(vault_cli_group, ["cook", "swh:1:dir:f00b4r", "-"])
    assert isinstance(result.exception, SystemExit)
    assert "expected core SWHID" in result.stdout

    result = runner.invoke(vault_cli_group, ["cook", "swh:1:ori:" + "0" * 40, "-"])
    assert isinstance(result.exception, SystemExit)
    assert "expected core SWHID" in result.stdout


def test_cook_unknown_cooker():
    runner = click.testing.CliRunner()

    result = runner.invoke(
        vault_cli_group,
        ["cook", "swh:1:dir:" + "0" * 40, "-", "--bundle-type", "gitfast"],
    )
    assert isinstance(result.exception, SystemExit)
    assert "do not have a gitfast cooker" in result.stdout

    result = runner.invoke(vault_cli_group, ["cook", "swh:1:rev:" + "0" * 40, "-"])
    assert isinstance(result.exception, SystemExit)
    assert "use --bundle-type" in result.stdout


@pytest.mark.parametrize(
    "bundle_type,cooker_name_suffix,swhid_type",
    [
        ("directory", "", "dir"),
        ("revision", "gitfast", "rev"),
    ],
)
def test_cook_directory(bundle_type, cooker_name_suffix, swhid_type, mocker):
    storage = object()
    mocker.patch("swh.storage.get_storage", return_value=storage)

    backend = MagicMock(spec=InMemoryVaultBackend)
    backend.fetch.return_value = b"bundle content"
    mocker.patch(
        "swh.vault.in_memory_backend.InMemoryVaultBackend", return_value=backend
    )

    cooker = MagicMock(spec=BaseVaultCooker)
    cooker_cls = MagicMock(return_value=cooker)
    mocker.patch("swh.vault.cookers.get_cooker_cls", return_value=cooker_cls)

    runner = click.testing.CliRunner()

    swhid = CoreSWHID.from_string(f"swh:1:{swhid_type}:{'0'*40}")

    with tempfile.NamedTemporaryFile("a", suffix=".yml") as config_fd:
        config_fd.write('{"storage": {}}')
        config_fd.seek(0)
        if cooker_name_suffix:
            result = runner.invoke(
                vault_cli_group,
                [
                    "cook",
                    f"swh:1:{swhid_type}:{'0'*40}",
                    "-",
                    "-C",
                    config_fd.name,
                    "--bundle-type",
                    cooker_name_suffix,
                ],
            )
        else:
            result = runner.invoke(
                vault_cli_group,
                ["cook", str(swhid), "-", "-C", config_fd.name],
            )

    if result.exception is not None:
        raise result.exception

    cooker_cls.assert_called_once_with(
        swhid=swhid,
        backend=backend,
        storage=storage,
        graph=None,
        objstorage=None,
        max_bundle_size=None,
    )
    cooker.cook.assert_called_once_with()

    assert result.stdout_bytes == b"bundle content"


def test_cli_swh_vault_db_create_and_init_db(postgresql, tmp_path):
    """Test that 'swh db init vault' works"""
    module_name = "vault"
    conninfo = craft_conninfo(postgresql, "new-db")

    cfgfile = tmp_path / "config.yml"
    CFG = f"""
vault:
  cls: postgresql
  db: {conninfo}
  cache:
    cls: memory
  storage:
    cls: memory
  scheduler:
    cls: remote
    url: mock://scheduler
    """
    cfgfile.write_text(CFG)

    cli_runner = click.testing.CliRunner()
    # This creates the db and installs the necessary admin extensions
    result = cli_runner.invoke(swhdb, ["create", module_name, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    result = cli_runner.invoke(swhdb, ["init-admin", module_name, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    # This initializes the schema and data
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "init", module_name])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    assert swh_db_module(conninfo) == "vault"
    assert swh_db_version(conninfo) == VaultBackend.current_version

    with BaseDb.connect(conninfo).cursor() as cur:
        cur.execute("select tablename from pg_tables where schemaname='public'")
        tables = {table for table, in cur.fetchall()}

    assert tables == {
        "dbmodule",
        "dbversion",
        "vault_bundle",
        "vault_notif_email",
        "vault_batch",
        "vault_batch_bundle",
    }



# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import contextlib
import datetime
import glob
import gzip
import io
import os
import pathlib
import shutil
import subprocess
import tarfile
import tempfile
import unittest
import unittest.mock

import attrs
import dulwich.fastexport
import dulwich.index
import dulwich.objects
import dulwich.porcelain
import dulwich.repo
import pytest

from swh.loader.git.from_disk import GitLoaderFromDisk
from swh.model import from_disk, hashutil
from swh.model.model import (
    Person,
    Release,
    Revision,
    RevisionType,
    SkippedContent,
    Snapshot,
    SnapshotBranch,
    TargetType,
    Timestamp,
    TimestampWithTimezone,
)
from swh.model.model import Content, Directory, DirectoryEntry
from swh.model.model import ObjectType as ModelObjectType
from swh.model.swhids import CoreSWHID, ObjectType
from swh.vault.cookers import DirectoryCooker, GitBareCooker, RevisionGitfastCooker
from swh.vault.tests.vault_testing import hash_content
from swh.vault.to_disk import HIDDEN_MESSAGE, SKIPPED_MESSAGE


class TestRepo:
    """A tiny context manager for a test git repository, with some utility
    functions to perform basic git stuff.
    """

    def __init__(self, repo_dir=None):
        self.repo_dir = repo_dir

    def __enter__(self):
        if self.repo_dir:
            self.tmp_dir = None
            self.repo = dulwich.repo.Repo(self.repo_dir)
        else:
            self.tmp_dir = tempfile.TemporaryDirectory(prefix="tmp-vault-repo-")
            self.repo_dir = self.tmp_dir.__enter__()
            self.repo = dulwich.repo.Repo.init(self.repo_dir)
        self.author_name = b"Test Author"
        self.author_email = b"test@softwareheritage.org"
        self.author = b"%s <%s>" % (self.author_name, self.author_email)
        self.base_date = 258244200
        self.counter = 0
        return pathlib.Path(self.repo_dir)

    def __exit__(self, exc, value, tb):
        if self.tmp_dir is not None:
            self.tmp_dir.__exit__(exc, value, tb)
            self.repo_dir = None

    def checkout(self, rev_sha):
        rev = self.repo[rev_sha]
        dulwich.index.build_index_from_tree(
            str(self.repo_dir), self.repo.index_path(), self.repo.object_store, rev.tree
        )

    def git_shell(self, *cmd, stdout=subprocess.DEVNULL, **kwargs):
        name = self.author_name
        email = self.author_email
        date = "%d +0000" % (self.base_date + self.counter)
        env = {
            # Set git commit format
            "GIT_AUTHOR_NAME": name,
            "GIT_AUTHOR_EMAIL": email,
            "GIT_AUTHOR_DATE": date,
            "GIT_COMMITTER_NAME": name,
            "GIT_COMMITTER_EMAIL": email,
            "GIT_COMMITTER_DATE": date,
            # Ignore all the system-wide and user configurations
            "GIT_CONFIG_NOSYSTEM": "1",
            "HOME": str(self.tmp_dir),
            "XDG_CONFIG_HOME": str(self.tmp_dir),
        }
        kwargs.setdefault("env", {}).update(env)

        subprocess.check_call(
            ("git", "-C", self.repo_dir) + cmd, stdout=stdout, **kwargs
        )

    def commit(self, message="Commit test\n", ref=b"HEAD"):
        """Commit the current working tree in a new commit with message on
        the branch 'ref'.

        At the end of the commit, the reference should stay the same
        and the index should be clean.

        """
        paths = [
            os.path.relpath(path, self.repo_dir)
            for path in glob.glob(self.repo_dir + "/**/*", recursive=True)
        ]
        self.repo.stage(paths)
        message = message.encode() + b"\n"
        ret = self.repo.do_commit(
            message=message,
            committer=self.author,
            commit_timestamp=self.base_date + self.counter,
            commit_timezone=0,
            ref=ref,
        )
        self.counter += 1

        # committing on another branch leaves
        # dangling files in index
        if ref != b"HEAD":
            # XXX this should work (but does not)
            # dulwich.porcelain.reset(self.repo, 'hard')
            self.git_shell("reset", "--hard", "HEAD")
        return ret

    def tag(self, name, target=b"HEAD", message=None):
        dulwich.porcelain.tag_create(
            self.repo,
            name,
            message=message,
            annotated=message is not None,
            objectish=target,
        )

    def merge(self, parent_sha_list, message="Merge branches."):
        self.git_shell(
            "merge",
            "--allow-unrelated-histories",
            "-m",
            message,
            *[p.decode() for p in parent_sha_list],
        )
        self.counter += 1
        return self.repo.refs[b"HEAD"]

    def print_debug_graph(self, reflog=False):
        args = ["log", "--all", "--graph", "--decorate"]
        if reflog:
            args.append("--reflog")
        self.git_shell(*args, stdout=None)


@pytest.fixture
def git_loader(
    swh_storage,
):
    """Instantiate a Git Loader using the storage instance as storage."""

    def _create_loader(directory):
        return GitLoaderFromDisk(
            swh_storage,
            "fake_origin",
            directory=directory,
            visit_date=datetime.datetime.now(datetime.timezone.utc),
        )

    return _create_loader


@contextlib.contextmanager
def cook_extract_directory_dircooker(storage, swhid, fsck=True):
    """Context manager that cooks a directory and extract it."""
    backend = unittest.mock.MagicMock()
    backend.storage = storage
    cooker = DirectoryCooker(swhid, backend=backend, storage=storage)
    cooker.fileobj = io.BytesIO()
    assert cooker.check_exists()
    cooker.prepare_bundle()
    cooker.fileobj.seek(0)
    with tempfile.TemporaryDirectory(prefix="tmp-vault-extract-") as td:
        with tarfile.open(fileobj=cooker.fileobj, mode="r") as tar:
            tar.extractall(td)
        yield pathlib.Path(td) / str(swhid)
    cooker.storage = None


@contextlib.contextmanager
def cook_extract_directory_gitfast(storage, swhid, fsck=True):
    """Context manager that cooks a revision containing a directory and extract it,
    using RevisionGitfastCooker"""
    test_repo = TestRepo()
    with test_repo as p:
        date = TimestampWithTimezone.from_datetime(
            datetime.datetime.now(datetime.timezone.utc)
        )
        revision = Revision(
            directory=swhid.object_id,
            message=b"dummy message",
            author=Person.from_fullname(b"someone"),
            committer=Person.from_fullname(b"someone"),
            date=date,
            committer_date=date,
            type=RevisionType.GIT,
            synthetic=False,
        )
        storage.revision_add([revision])

    with cook_stream_revision_gitfast(
        storage, revision.swhid()
    ) as stream, test_repo as p:
        processor = dulwich.fastexport.GitImportProcessor(test_repo.repo)
        processor.import_stream(stream)
        test_repo.checkout(b"HEAD")
        shutil.rmtree(p / ".git")
        yield p


@contextlib.contextmanager
def cook_extract_directory_git_bare(storage, swhid, fsck=True, direct_objstorage=False):
    """Context manager that cooks a revision and extract it,
    using GitBareCooker"""
    backend = unittest.mock.MagicMock()
    backend.storage = storage

    # Cook the object
    cooker = GitBareCooker(
        swhid,
        backend=backend,
        storage=storage,
        objstorage=storage.objstorage if direct_objstorage else None,
    )
    cooker.use_fsck = fsck  # Some tests try edge-cases that git-fsck rejects
    cooker.fileobj = io.BytesIO()
    assert cooker.check_exists()
    cooker.prepare_bundle()
    cooker.fileobj.seek(0)

    # Extract it
    with tempfile.TemporaryDirectory(prefix="tmp-vault-extract-") as td:
        with tarfile.open(fileobj=cooker.fileobj, mode="r") as tar:
            tar.extractall(td)

        # Clone it with Dulwich
        with tempfile.TemporaryDirectory(prefix="tmp-vault-clone-") as clone_dir:
            clone_dir = pathlib.Path(clone_dir)
            subprocess.check_call(
                [
                    "git",
                    "clone",
                    os.path.join(td, f"{swhid}.git"),
                    clone_dir,
                ]
            )
            shutil.rmtree(clone_dir / ".git")
            yield clone_dir


@pytest.fixture(
    scope="module",
    params=[
        cook_extract_directory_dircooker,
        cook_extract_directory_gitfast,
        cook_extract_directory_git_bare,
    ],
)
def cook_extract_directory(request):
    """A fixture that is instantiated as either cook_extract_directory_dircooker or
    cook_extract_directory_git_bare."""
    return request.param


@contextlib.contextmanager
def cook_stream_revision_gitfast(storage, swhid):
    """Context manager that cooks a revision and stream its fastexport."""
    backend = unittest.mock.MagicMock()
    backend.storage = storage
    cooker = RevisionGitfastCooker(swhid, backend=backend, storage=storage)
    cooker.fileobj = io.BytesIO()
    assert cooker.check_exists()
    cooker.prepare_bundle()
    cooker.fileobj.seek(0)
    fastexport_stream = gzip.GzipFile(fileobj=cooker.fileobj)
    yield fastexport_stream
    cooker.storage = None


@contextlib.contextmanager
def cook_extract_revision_gitfast(storage, swhid, fsck=True):
    """Context manager that cooks a revision and extract it,
    using RevisionGitfastCooker"""
    test_repo = TestRepo()
    with cook_stream_revision_gitfast(storage, swhid) as stream, test_repo as p:
        processor = dulwich.fastexport.GitImportProcessor(test_repo.repo)
        processor.import_stream(stream)
        yield test_repo, p


@contextlib.contextmanager
def cook_extract_git_bare(storage, swhid, fsck=True):
    """Context manager that cooks a revision and extract it,
    using GitBareCooker"""
    backend = unittest.mock.MagicMock()
    backend.storage = storage

    # Cook the object
    cooker = GitBareCooker(swhid, backend=backend, storage=storage)
    cooker.use_fsck = fsck  # Some tests try edge-cases that git-fsck rejects
    cooker.fileobj = io.BytesIO()
    assert cooker.check_exists()
    cooker.prepare_bundle()
    cooker.fileobj.seek(0)

    # Extract it
    with tempfile.TemporaryDirectory(prefix="tmp-vault-extract-") as td:
        with tarfile.open(fileobj=cooker.fileobj, mode="r") as tar:
            tar.extractall(td)

        # Clone it with Dulwich
        with tempfile.TemporaryDirectory(prefix="tmp-vault-clone-") as clone_dir:
            clone_dir = pathlib.Path(clone_dir)
            subprocess.check_call(
                [
                    "git",
                    "clone",
                    os.path.join(td, f"{swhid}.git"),
                    clone_dir,
                ]
            )
            test_repo = TestRepo(clone_dir)
            with test_repo:
                yield test_repo, clone_dir


@contextlib.contextmanager
def cook_extract_revision_git_bare(storage, swhid, fsck=True):
    with cook_extract_git_bare(
        storage,
        swhid,
        fsck=fsck,
    ) as res:
        yield res


@pytest.fixture(
    scope="module",
    params=[cook_extract_revision_gitfast, cook_extract_revision_git_bare],
)
def cook_extract_revision(request):
    """A fixture that is instantiated as either cook_extract_revision_gitfast or
    cook_extract_revision_git_bare."""
    return request.param


@contextlib.contextmanager
def cook_extract_snapshot_git_bare(storage, swhid, fsck=True):
    with cook_extract_git_bare(
        storage,
        swhid,
        fsck=fsck,
    ) as res:
        yield res


@pytest.fixture(
    scope="module",
    params=[cook_extract_snapshot_git_bare],
)
def cook_extract_snapshot(request):
    """Equivalent to cook_extract_snapshot_git_bare; but analogous to
    cook_extract_revision in case we ever have more cookers supporting snapshots"""
    return request.param


TEST_CONTENT = (
    "   test content\n" "and unicode \N{BLACK HEART SUIT}\n" " and trailing spaces   "
)
TEST_EXECUTABLE = b"\x42\x40\x00\x00\x05"


class TestDirectoryCooker:
    def test_directory_simple(self, git_loader, cook_extract_directory):
        repo = TestRepo()
        with repo as rp:
            (rp / "file").write_text(TEST_CONTENT)
            (rp / "executable").write_bytes(TEST_EXECUTABLE)
            (rp / "executable").chmod(0o755)
            (rp / "link").symlink_to("file")
            (rp / "dir1/dir2").mkdir(parents=True)
            (rp / "dir1/dir2/file").write_text(TEST_CONTENT)
            c = repo.commit()
            loader = git_loader(str(rp))
            loader.load()

            obj_id_hex = repo.repo[c].tree.decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)
            swhid = CoreSWHID(object_type=ObjectType.DIRECTORY, object_id=obj_id)

        with cook_extract_directory(loader.storage, swhid) as p:
            assert (p / "file").stat().st_mode == 0o100644
            assert (p / "file").read_text() == TEST_CONTENT
            assert (p / "executable").stat().st_mode == 0o100755
            assert (p / "executable").read_bytes() == TEST_EXECUTABLE
            assert (p / "link").is_symlink()
            assert os.readlink(str(p / "link")) == "file"
            assert (p / "dir1/dir2/file").stat().st_mode == 0o100644
            assert (p / "dir1/dir2/file").read_text() == TEST_CONTENT

            directory = from_disk.Directory.from_disk(path=bytes(p))
            assert obj_id_hex == hashutil.hash_to_hex(directory.hash)

    def test_directory_filtered_objects(self, git_loader, cook_extract_directory):
        repo = TestRepo()
        with repo as rp:
            file_1, id_1 = hash_content(b"test1")
            file_2, id_2 = hash_content(b"test2")
            file_3, id_3 = hash_content(b"test3")

            (rp / "file").write_bytes(file_1)
            (rp / "hidden_file").write_bytes(file_2)
            (rp / "absent_file").write_bytes(file_3)

            c = repo.commit()
            loader = git_loader(str(rp))
            loader.load()

            obj_id_hex = repo.repo[c].tree.decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)
            swhid = CoreSWHID(object_type=ObjectType.DIRECTORY, object_id=obj_id)

        # alter the content of the storage
        # 1/ make file 2 an hidden file object
        loader.storage._allow_overwrite = True
        cnt2 = attrs.evolve(
            loader.storage.content_get([id_2])[0], status="hidden", data=file_2
        )
        loader.storage.content_add([cnt2])
        assert loader.storage.content_get([id_2])[0].status == "hidden"

        # 2/ make file 3 an skipped file object
        cnt3 = loader.storage.content_get([id_3])[0].to_dict()
        cnt3["status"] = "absent"
        cnt3["reason"] = "no reason"
        sk_cnt3 = SkippedContent.from_dict(cnt3)
        loader.storage.skipped_content_add([sk_cnt3])
        # dirty dirty dirty... let's pretend it is the equivalent of writing sql
        # queries in the postgresql backend
        for hashkey in loader.storage._cql_runner._content_indexes:
            loader.storage._cql_runner._content_indexes[hashkey].pop(cnt3[hashkey])

        with cook_extract_directory(loader.storage, swhid) as p:
            assert (p / "file").read_bytes() == b"test1"
            assert (p / "hidden_file").read_bytes() == HIDDEN_MESSAGE
            assert (p / "absent_file").read_bytes() == SKIPPED_MESSAGE

    def test_directory_bogus_perms(self, git_loader, cook_extract_directory):
        # Some early git repositories have 664/775 permissions... let's check
        # if all the weird modes are properly normalized in the directory
        # cooker.
        repo = TestRepo()
        with repo as rp:
            (rp / "file").write_text(TEST_CONTENT)
            (rp / "file").chmod(0o664)
            (rp / "executable").write_bytes(TEST_EXECUTABLE)
            (rp / "executable").chmod(0o775)
            (rp / "wat").write_text(TEST_CONTENT)
            (rp / "wat").chmod(0o604)

            # Disable mode cleanup
            with unittest.mock.patch("dulwich.index.cleanup_mode", lambda mode: mode):
                c = repo.commit()

            # Make sure Dulwich didn't normalize the permissions itself.
            # (if it did, then the test can't check the cooker normalized them)
            tree_id = repo.repo[c].tree
            assert {entry.mode for entry in repo.repo[tree_id].items()} == {
                0o100775,
                0o100664,
                0o100604,
            }

            # Disable mode checks
            with unittest.mock.patch("dulwich.objects.Tree.check", lambda self: None):
                loader = git_loader(str(rp))
                loader.load()

            # Make sure swh-loader didn't normalize them either
            dir_entries = loader.storage.directory_ls(hashutil.bytehex_to_hash(tree_id))
            assert {entry["perms"] for entry in dir_entries} == {
                0o100664,
                0o100775,
                0o100604,
            }

            obj_id_hex = repo.repo[c].tree.decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)
            swhid = CoreSWHID(object_type=ObjectType.DIRECTORY, object_id=obj_id)

        with cook_extract_directory(loader.storage, swhid) as p:
            assert (p / "file").stat().st_mode == 0o100644
            assert (p / "executable").stat().st_mode == 0o100755
            assert (p / "wat").stat().st_mode == 0o100644

    @pytest.mark.parametrize("direct_objstorage", [True, False])
    def test_directory_objstorage(
        self, swh_storage, git_loader, mocker, direct_objstorage
    ):
        """Like test_directory_simple, but using swh_objstorage directly, without
        going through swh_storage.content_get_data()"""
        repo = TestRepo()
        with repo as rp:
            (rp / "file").write_text(TEST_CONTENT)
            (rp / "executable").write_bytes(TEST_EXECUTABLE)
            (rp / "executable").chmod(0o755)
            (rp / "link").symlink_to("file")
            (rp / "dir1/dir2").mkdir(parents=True)
            (rp / "dir1/dir2/file").write_text(TEST_CONTENT)
            c = repo.commit()
            loader = git_loader(str(rp))
            loader.load()

            obj_id_hex = repo.repo[c].tree.decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)
            swhid = CoreSWHID(object_type=ObjectType.DIRECTORY, object_id=obj_id)

        # Set-up spies
        storage_content_get_data = mocker.patch.object(
            swh_storage, "content_get_data", wraps=swh_storage.content_get_data
        )
        objstorage_content_batch = mocker.patch.object(
            swh_storage.objstorage, "get_batch", wraps=swh_storage.objstorage.get_batch
        )

        with cook_extract_directory_git_bare(
            loader.storage, swhid, direct_objstorage=direct_objstorage
        ) as p:
            assert (p / "file").stat().st_mode == 0o100644
            assert (p / "file").read_text() == TEST_CONTENT
            assert (p / "executable").stat().st_mode == 0o100755
            assert (p / "executable").read_bytes() == TEST_EXECUTABLE
            assert (p / "link").is_symlink()
            assert os.readlink(str(p / "link")) == "file"
            assert (p / "dir1/dir2/file").stat().st_mode == 0o100644
            assert (p / "dir1/dir2/file").read_text() == TEST_CONTENT

            directory = from_disk.Directory.from_disk(path=bytes(p))
            assert obj_id_hex == hashutil.hash_to_hex(directory.hash)

        if direct_objstorage:
            storage_content_get_data.assert_not_called()
            objstorage_content_batch.assert_called()
        else:
            storage_content_get_data.assert_called()
            objstorage_content_batch.assert_not_called()

    def test_directory_revision_data(self, swh_storage):
        target_rev = "0e8a3ad980ec179856012b7eecf4327e99cd44cd"

        dir = Directory(
            entries=(
                DirectoryEntry(
                    name=b"submodule",
                    type="rev",
                    target=hashutil.hash_to_bytes(target_rev),
                    perms=0o100644,
                ),
            ),
        )
        swh_storage.directory_add([dir])

        with cook_extract_directory_dircooker(
            swh_storage, dir.swhid(), fsck=False
        ) as p:
            assert (p / "submodule").is_symlink()
            assert os.readlink(str(p / "submodule")) == target_rev


class RepoFixtures:
    """Shared loading and checking methods that can be reused by different types
    of tests."""

    def load_repo_simple(self, git_loader):
        #
        #     1--2--3--4--5--6--7
        #
        repo = TestRepo()
        with repo as rp:
            (rp / "file1").write_text(TEST_CONTENT)
            repo.commit("add file1")
            (rp / "file2").write_text(TEST_CONTENT)
            repo.commit("add file2")
            (rp / "dir1/dir2").mkdir(parents=True)
            (rp / "dir1/dir2/file").write_text(TEST_CONTENT)

            (rp / "bin1").write_bytes(TEST_EXECUTABLE)
            (rp / "bin1").chmod(0o755)
            repo.commit("add bin1")
            (rp / "link1").symlink_to("file1")
            repo.commit("link link1 to file1")
            (rp / "file2").unlink()
            repo.commit("remove file2")
            (rp / "bin1").rename(rp / "bin")
            repo.commit("rename bin1 to bin")
            loader = git_loader(str(rp))
            loader.load()
            obj_id_hex = repo.repo.refs[b"HEAD"].decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)
            swhid = CoreSWHID(object_type=ObjectType.REVISION, object_id=obj_id)
        return (loader, swhid)

    def check_revision_simple(self, ert, p, swhid):
        ert.checkout(b"HEAD")
        assert (p / "file1").stat().st_mode == 0o100644
        assert (p / "file1").read_text() == TEST_CONTENT
        assert (p / "link1").is_symlink()
        assert os.readlink(str(p / "link1")) == "file1"
        assert (p / "bin").stat().st_mode == 0o100755
        assert (p / "bin").read_bytes() == TEST_EXECUTABLE
        assert (p / "dir1/dir2/file").read_text() == TEST_CONTENT
        assert (p / "dir1/dir2/file").stat().st_mode == 0o100644
        assert ert.repo.refs[b"HEAD"].decode() == swhid.object_id.hex()

    def load_repo_two_roots(self, git_loader):
        #
        #    1----3---4
        #        /
        #   2----
        #
        repo = TestRepo()
        with repo as rp:
            (rp / "file1").write_text(TEST_CONTENT)
            c1 = repo.commit("Add file1")
            del repo.repo.refs[b"refs/heads/master"]  # git update-ref -d HEAD
            (rp / "file2").write_text(TEST_CONTENT)
            repo.commit("Add file2")
            repo.merge([c1])
            (rp / "file3").write_text(TEST_CONTENT)
            repo.commit("add file3")
            obj_id_hex = repo.repo.refs[b"HEAD"].decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)
            swhid = CoreSWHID(object_type=ObjectType.REVISION, object_id=obj_id)
            loader = git_loader(str(rp))
            loader.load()
        return (loader, swhid)

    def check_revision_two_roots(self, ert, p, swhid):
        assert ert.repo.refs[b"HEAD"].decode() == swhid.object_id.hex()

        (c3,) = ert.repo[hashutil.hash_to_bytehex(swhid.object_id)].parents
        assert len(ert.repo[c3].parents) == 2

    def load_repo_two_heads(self, git_loader):
        #
        #    1---2----4      <-- master and b1
        #         \
        #          ----3     <-- b2
        #
        repo = TestRepo()
        with repo as rp:
            (rp / "file1").write_text(TEST_CONTENT)
            repo.commit("Add file1")

            (rp / "file2").write_text(TEST_CONTENT)
            c2 = repo.commit("Add file2")

            repo.repo.refs[b"refs/heads/b2"] = c2  # branch b2 from master

            (rp / "file3").write_text(TEST_CONTENT)
            repo.commit("add file3", ref=b"refs/heads/b2")

            (rp / "file4").write_text(TEST_CONTENT)
            c4 = repo.commit("add file4", ref=b"refs/heads/master")
            repo.repo.refs[b"refs/heads/b1"] = c4  # branch b1 from master

            obj_id_hex = repo.repo.refs[b"HEAD"].decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)
            swhid = CoreSWHID(object_type=ObjectType.REVISION, object_id=obj_id)
            loader = git_loader(str(rp))
            loader.load()
        return (loader, swhid)

    def check_snapshot_two_heads(self, ert, p, swhid):
        assert (
            hashutil.hash_to_bytehex(swhid.object_id)
            == ert.repo.refs[b"HEAD"]
            == ert.repo.refs[b"refs/heads/master"]
            == ert.repo.refs[b"refs/remotes/origin/HEAD"]
            == ert.repo.refs[b"refs/remotes/origin/master"]
            == ert.repo.refs[b"refs/remotes/origin/b1"]
        )

        c4_id = hashutil.hash_to_bytehex(swhid.object_id)
        c3_id = ert.repo.refs[b"refs/remotes/origin/b2"]

        assert ert.repo[c3_id].parents == ert.repo[c4_id].parents

    def load_repo_two_double_fork_merge(self, git_loader):
        #
        #     2---4---6
        #    /   /   /
        #   1---3---5
        #
        repo = TestRepo()
        with repo as rp:
            (rp / "file1").write_text(TEST_CONTENT)
            c1 = repo.commit("Add file1")  # create commit 1
            repo.repo.refs[b"refs/heads/c1"] = c1  # branch c1 from master

            (rp / "file2").write_text(TEST_CONTENT)
            repo.commit("Add file2")  # create commit 2

            (rp / "file3").write_text(TEST_CONTENT)
            c3 = repo.commit("Add file3", ref=b"refs/heads/c1")  # create commit 3 on c1
            repo.repo.refs[b"refs/heads/c3"] = c3  # branch c3 from c1

            repo.merge([c3])  # create commit 4

            (rp / "file5").write_text(TEST_CONTENT)
            c5 = repo.commit("Add file3", ref=b"refs/heads/c3")  # create commit 5 on c3

            repo.merge([c5])  # create commit 6

            obj_id_hex = repo.repo.refs[b"HEAD"].decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)
            swhid = CoreSWHID(object_type=ObjectType.REVISION, object_id=obj_id)
            loader = git_loader(str(rp))
            loader.load()
        return (loader, swhid)

    def check_revision_two_double_fork_merge(self, ert, p, swhid):
        assert ert.repo.refs[b"HEAD"].decode() == swhid.object_id.hex()

    def check_snapshot_two_double_fork_merge(self, ert, p, swhid):
        assert (
            hashutil.hash_to_bytehex(swhid.object_id)
            == ert.repo.refs[b"HEAD"]
            == ert.repo.refs[b"refs/heads/master"]
            == ert.repo.refs[b"refs/remotes/origin/HEAD"]
            == ert.repo.refs[b"refs/remotes/origin/master"]
        )

        (c4_id, c5_id) = ert.repo[swhid.object_id.hex().encode()].parents
        assert c5_id == ert.repo.refs[b"refs/remotes/origin/c3"]

        (c2_id, c3_id) = ert.repo[c4_id].parents
        assert c3_id == ert.repo.refs[b"refs/remotes/origin/c1"]

    def load_repo_triple_merge(self, git_loader):
        #
        #       .---.---5
        #      /   /   /
        #     2   3   4
        #    /   /   /
        #   1---.---.
        #
        repo = TestRepo()
        with repo as rp:
            (rp / "file1").write_text(TEST_CONTENT)
            c1 = repo.commit("Commit 1")
            repo.repo.refs[b"refs/heads/b1"] = c1
            repo.repo.refs[b"refs/heads/b2"] = c1

            repo.commit("Commit 2")
            c3 = repo.commit("Commit 3", ref=b"refs/heads/b1")
            c4 = repo.commit("Commit 4", ref=b"refs/heads/b2")
            repo.merge([c3, c4])

            obj_id_hex = repo.repo.refs[b"HEAD"].decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)
            swhid = CoreSWHID(object_type=ObjectType.REVISION, object_id=obj_id)
            loader = git_loader(str(rp))
            loader.load()
        return (loader, swhid)

    def check_revision_triple_merge(self, ert, p, swhid):
        assert ert.repo.refs[b"HEAD"].decode() == swhid.object_id.hex()

    def check_snapshot_triple_merge(self, ert, p, swhid):
        assert (
            hashutil.hash_to_bytehex(swhid.object_id)
            == ert.repo.refs[b"HEAD"]
            == ert.repo.refs[b"refs/heads/master"]
            == ert.repo.refs[b"refs/remotes/origin/HEAD"]
            == ert.repo.refs[b"refs/remotes/origin/master"]
        )

        (c2_id, c3_id, c4_id) = ert.repo[swhid.object_id.hex().encode()].parents
        assert c3_id == ert.repo.refs[b"refs/remotes/origin/b1"]
        assert c4_id == ert.repo.refs[b"refs/remotes/origin/b2"]

        assert (
            ert.repo[c2_id].parents
            == ert.repo[c3_id].parents
            == ert.repo[c4_id].parents
        )

    def load_repo_filtered_objects(self, git_loader):
        repo = TestRepo()
        with repo as rp:
            file_1, id_1 = hash_content(b"test1")
            file_2, id_2 = hash_content(b"test2")
            file_3, id_3 = hash_content(b"test3")

            (rp / "file").write_bytes(file_1)
            (rp / "hidden_file").write_bytes(file_2)
            (rp / "absent_file").write_bytes(file_3)

            repo.commit()
            obj_id_hex = repo.repo.refs[b"HEAD"].decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)
            swhid = CoreSWHID(object_type=ObjectType.REVISION, object_id=obj_id)
            loader = git_loader(str(rp))
            loader.load()

        # alter the content of the storage
        # 1/ make file 2 an hidden file object
        loader.storage._allow_overwrite = True
        cnt2 = attrs.evolve(
            loader.storage.content_get([id_2])[0], status="hidden", data=file_2
        )
        loader.storage.content_add([cnt2])
        assert loader.storage.content_get([id_2])[0].status == "hidden"

        # 2/ make file 3 an skipped file object
        cnt3 = loader.storage.content_get([id_3])[0].to_dict()
        cnt3["status"] = "absent"
        cnt3["reason"] = "no reason"
        sk_cnt3 = SkippedContent.from_dict(cnt3)
        loader.storage.skipped_content_add([sk_cnt3])
        # dirty dirty dirty... let's pretend it is the equivalent of writing sql
        # queries in the postgresql backend
        for hashkey in loader.storage._cql_runner._content_indexes:
            loader.storage._cql_runner._content_indexes[hashkey].pop(cnt3[hashkey])

        return (loader, swhid)

    def check_revision_filtered_objects(self, ert, p, swhid):
        ert.checkout(b"HEAD")
        assert (p / "file").read_bytes() == b"test1"
        assert (p / "hidden_file").read_bytes() == HIDDEN_MESSAGE
        assert (p / "absent_file").read_bytes() == SKIPPED_MESSAGE

    def load_repo_null_fields(self, git_loader):
        # Our schema doesn't enforce a lot of non-null revision fields. We need
        # to check these cases don't break the cooker.
        repo = TestRepo()
        with repo as rp:
            (rp / "file").write_text(TEST_CONTENT)
            c = repo.commit("initial commit")
            loader = git_loader(str(rp))
            loader.load()
            repo.repo.refs[b"HEAD"].decode()
            dir_id_hex = repo.repo[c].tree.decode()
            dir_id = hashutil.hash_to_bytes(dir_id_hex)

        test_revision = Revision(
            message=b"",
            author=Person(name=None, email=None, fullname=b""),
            date=None,
            committer=Person(name=None, email=None, fullname=b""),
            committer_date=None,
            parents=(),
            type=RevisionType.GIT,
            directory=dir_id,
            metadata={},
            synthetic=True,
        )

        storage = loader.storage
        storage.revision_add([test_revision])
        return (loader, test_revision.swhid())

    def check_revision_null_fields(self, ert, p, swhid):
        ert.checkout(b"HEAD")
        assert (p / "file").stat().st_mode == 0o100644

    def load_repo_tags(self, git_loader):
        #        v-- t2
        #
        #    1---2----5      <-- master, t5, and t5a (annotated)
        #         \
        #          ----3----4     <-- t4a (annotated)
        #
        repo = TestRepo()
        with repo as rp:
            (rp / "file1").write_text(TEST_CONTENT)
            repo.commit("Add file1")

            (rp / "file2").write_text(TEST_CONTENT)
            repo.commit("Add file2")  # create c2

            repo.tag(b"t2")

            (rp / "file3").write_text(TEST_CONTENT)
            repo.commit("add file3")

            (rp / "file4").write_text(TEST_CONTENT)
            repo.commit("add file4")

            repo.tag(b"t4a", message=b"tag 4")

            # Go back to c2
            repo.git_shell("reset", "--hard", "HEAD^^")

            (rp / "file5").write_text(TEST_CONTENT)
            repo.commit("add file5")  # create c5

            repo.tag(b"t5")
            repo.tag(b"t5a", message=b"tag 5")

            obj_id_hex = repo.repo.refs[b"HEAD"].decode()
            obj_id = hashutil.hash_to_bytes(obj_id_hex)
            swhid = CoreSWHID(object_type=ObjectType.REVISION, object_id=obj_id)
            loader = git_loader(str(rp))
            loader.load()
        return (loader, swhid)

    def check_snapshot_tags(self, ert, p, swhid):
        assert (
            hashutil.hash_to_bytehex(swhid.object_id)
            == ert.repo.refs[b"HEAD"]
            == ert.repo.refs[b"refs/heads/master"]
            == ert.repo.refs[b"refs/remotes/origin/HEAD"]
            == ert.repo.refs[b"refs/remotes/origin/master"]
            == ert.repo.refs[b"refs/tags/t5"]
        )

        c2_id = ert.repo.refs[b"refs/tags/t2"]
        c5_id = hashutil.hash_to_bytehex(swhid.object_id)

        assert ert.repo[c5_id].parents == [c2_id]

        t5a = ert.repo[ert.repo.refs[b"refs/tags/t5a"]]
        # TODO: investigate why new dulwich adds \n
        assert t5a.message in (b"tag 5", b"tag 5\n")
        assert t5a.object == (dulwich.objects.Commit, c5_id)

        t4a = ert.repo[ert.repo.refs[b"refs/tags/t4a"]]
        (_, c4_id) = t4a.object
        assert ert.repo[c4_id].message == b"add file4\n"  # TODO: ditto
        (c3_id,) = ert.repo[c4_id].parents
        assert ert.repo[c3_id].message == b"add file3\n"  # TODO: ditto
        assert ert.repo[c3_id].parents == [c2_id]


class TestRevisionCooker(RepoFixtures):
    def test_revision_simple(self, git_loader, cook_extract_revision):
        (loader, swhid) = self.load_repo_simple(git_loader)
        with cook_extract_revision(loader.storage, swhid) as (ert, p):
            self.check_revision_simple(ert, p, swhid)

    def test_revision_two_roots(self, git_loader, cook_extract_revision):
        (loader, swhid) = self.load_repo_two_roots(git_loader)
        with cook_extract_revision(loader.storage, swhid) as (ert, p):
            self.check_revision_two_roots(ert, p, swhid)

    def test_revision_two_double_fork_merge(self, git_loader, cook_extract_revision):
        (loader, swhid) = self.load_repo_two_double_fork_merge(git_loader)
        with cook_extract_revision(loader.storage, swhid) as (ert, p):
            self.check_revision_two_double_fork_merge(ert, p, swhid)

    def test_revision_triple_merge(self, git_loader, cook_extract_revision):
        (loader, swhid) = self.load_repo_triple_merge(git_loader)
        with cook_extract_revision(loader.storage, swhid) as (ert, p):
            self.check_revision_triple_merge(ert, p, swhid)

    def test_revision_filtered_objects(self, git_loader, cook_extract_revision):
        (loader, swhid) = self.load_repo_filtered_objects(git_loader)
        with cook_extract_revision(loader.storage, swhid) as (ert, p):
            self.check_revision_filtered_objects(ert, p, swhid)

    def test_revision_null_fields(self, git_loader, cook_extract_revision):
        (loader, swhid) = self.load_repo_null_fields(git_loader)
        with cook_extract_revision(loader.storage, swhid, fsck=False) as (ert, p):
            self.check_revision_null_fields(ert, p, swhid)

    @pytest.mark.parametrize("ingest_target_revision", [False, True])
    def test_revision_submodule(
        self, swh_storage, cook_extract_revision, ingest_target_revision
    ):
        date = TimestampWithTimezone.from_datetime(
            datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
        )

        target_rev = Revision(
            message=b"target_rev",
            author=Person.from_fullname(b"me <test@example.org>"),
            date=date,
            committer=Person.from_fullname(b"me <test@example.org>"),
            committer_date=date,
            parents=(),
            type=RevisionType.GIT,
            directory=bytes.fromhex("3333333333333333333333333333333333333333"),
            metadata={},
            synthetic=True,
        )
        if ingest_target_revision:
            swh_storage.revision_add([target_rev])

        dir = Directory(
            entries=(
                DirectoryEntry(
                    name=b"submodule",
                    type="rev",
                    target=target_rev.id,
                    perms=0o160000,
                ),
            ),
        )
        swh_storage.directory_add([dir])

        rev = Revision(
            message=b"msg",
            author=Person.from_fullname(b"me <test@example.org>"),
            date=date,
            committer=Person.from_fullname(b"me <test@example.org>"),
            committer_date=date,
            parents=(),
            type=RevisionType.GIT,
            directory=dir.id,
            metadata={},
            synthetic=True,
        )
        swh_storage.revision_add([rev])

        with cook_extract_revision(swh_storage, rev.swhid()) as (ert, p):
            ert.checkout(b"HEAD")
            pattern = b"160000 submodule\x00%s" % target_rev.id
            tree = ert.repo[b"HEAD"].tree
            assert pattern in ert.repo[tree].as_raw_string()


class TestSnapshotCooker(RepoFixtures):
    def test_snapshot_simple(self, git_loader, cook_extract_snapshot):
        (loader, main_rev_id) = self.load_repo_simple(git_loader)
        snp_id = loader.loaded_snapshot_id
        swhid = CoreSWHID(object_type=ObjectType.SNAPSHOT, object_id=snp_id)
        with cook_extract_snapshot(loader.storage, swhid) as (ert, p):
            self.check_revision_simple(ert, p, main_rev_id)

    def test_snapshot_two_roots(self, git_loader, cook_extract_snapshot):
        (loader, main_rev_id) = self.load_repo_two_roots(git_loader)
        snp_id = loader.loaded_snapshot_id
        swhid = CoreSWHID(object_type=ObjectType.SNAPSHOT, object_id=snp_id)
        with cook_extract_snapshot(loader.storage, swhid) as (ert, p):
            self.check_revision_two_roots(ert, p, main_rev_id)

    def test_snapshot_two_heads(self, git_loader, cook_extract_snapshot):
        (loader, main_rev_id) = self.load_repo_two_heads(git_loader)
        snp_id = loader.loaded_snapshot_id
        swhid = CoreSWHID(object_type=ObjectType.SNAPSHOT, object_id=snp_id)
        with cook_extract_snapshot(loader.storage, swhid) as (ert, p):
            self.check_snapshot_two_heads(ert, p, main_rev_id)

    def test_snapshot_two_double_fork_merge(self, git_loader, cook_extract_snapshot):
        (loader, main_rev_id) = self.load_repo_two_double_fork_merge(git_loader)
        snp_id = loader.loaded_snapshot_id
        swhid = CoreSWHID(object_type=ObjectType.SNAPSHOT, object_id=snp_id)
        with cook_extract_snapshot(loader.storage, swhid) as (ert, p):
            self.check_revision_two_double_fork_merge(ert, p, main_rev_id)
            self.check_snapshot_two_double_fork_merge(ert, p, main_rev_id)

    def test_snapshot_triple_merge(self, git_loader, cook_extract_snapshot):
        (loader, main_rev_id) = self.load_repo_triple_merge(git_loader)
        snp_id = loader.loaded_snapshot_id
        swhid = CoreSWHID(object_type=ObjectType.SNAPSHOT, object_id=snp_id)
        with cook_extract_snapshot(loader.storage, swhid) as (ert, p):
            self.check_revision_triple_merge(ert, p, main_rev_id)
            self.check_snapshot_triple_merge(ert, p, main_rev_id)

    def test_snapshot_filtered_objects(self, git_loader, cook_extract_snapshot):
        (loader, main_rev_id) = self.load_repo_filtered_objects(git_loader)
        snp_id = loader.loaded_snapshot_id
        swhid = CoreSWHID(object_type=ObjectType.SNAPSHOT, object_id=snp_id)
        with cook_extract_snapshot(loader.storage, swhid) as (ert, p):
            self.check_revision_filtered_objects(ert, p, main_rev_id)

    def test_snapshot_tags(self, git_loader, cook_extract_snapshot):
        (loader, main_rev_id) = self.load_repo_tags(git_loader)
        snp_id = loader.loaded_snapshot_id
        swhid = CoreSWHID(object_type=ObjectType.SNAPSHOT, object_id=snp_id)
        with cook_extract_snapshot(loader.storage, swhid) as (ert, p):
            self.check_snapshot_tags(ert, p, main_rev_id)

    def test_original_malformed_objects(self, swh_storage, cook_extract_snapshot):
        """Tests that objects that were originally malformed:

        * are still interpreted somewhat correctly (if the loader could make sense of
          them), especially that they still have links to children
        * have their original manifest in the bundle
        """
        date = TimestampWithTimezone.from_numeric_offset(
            Timestamp(1643819927, 0), 0, False
        )

        content = Content.from_data(b"foo")
        swh_storage.content_add([content])

        # disordered
        # fmt: off
        malformed_dir_manifest = (
            b""
            + b"100644 file2\x00" + content.sha1_git
            + b"100644 file1\x00" + content.sha1_git
        )
        # fmt: on
        directory = Directory(
            entries=(
                DirectoryEntry(
                    name=b"file1", type="file", perms=0o100644, target=content.sha1_git
                ),
                DirectoryEntry(
                    name=b"file2", type="file", perms=0o100644, target=content.sha1_git
                ),
            ),
            raw_manifest=f"tree {len(malformed_dir_manifest)}\x00".encode()
            + malformed_dir_manifest,
        )
        swh_storage.directory_add([directory])

        # 'committer' and 'author' swapped
        # fmt: off
        malformed_rev_manifest = (
            b"tree " + hashutil.hash_to_bytehex(directory.id) + b"\n"
            + b"committer me <test@example.org> 1643819927 +0000\n"
            + b"author me <test@example.org> 1643819927 +0000\n"
            + b"\n"
            + b"rev"
        )
        # fmt: on
        revision = Revision(
            message=b"rev",
            author=Person.from_fullname(b"me <test@example.org>"),
            date=date,
            committer=Person.from_fullname(b"me <test@example.org>"),
            committer_date=date,
            parents=(),
            type=RevisionType.GIT,
            directory=directory.id,
            synthetic=True,
            raw_manifest=f"commit {len(malformed_rev_manifest)}\x00".encode()
            + malformed_rev_manifest,
        )
        swh_storage.revision_add([revision])

        # 'tag' and 'tagger' swapped
        # fmt: off
        malformed_rel_manifest = (
            b"object " + hashutil.hash_to_bytehex(revision.id) + b"\n"
            + b"type commit\n"
            + b"tagger me <test@example.org> 1643819927 +0000\n"
            + b"tag v1.1.0\n"
        )
        # fmt: on

        release = Release(
            name=b"v1.1.0",
            message=None,
            author=Person.from_fullname(b"me <test@example.org>"),
            date=date,
            target=revision.id,
            target_type=ModelObjectType.REVISION,
            synthetic=True,
            raw_manifest=f"tag {len(malformed_rel_manifest)}\x00".encode()
            + malformed_rel_manifest,
        )
        swh_storage.release_add([release])

        snapshot = Snapshot(
            branches={
                b"refs/tags/v1.1.0": SnapshotBranch(
                    target=release.id, target_type=TargetType.RELEASE
                ),
                b"HEAD": SnapshotBranch(
                    target=revision.id, target_type=TargetType.REVISION
                ),
            }
        )
        swh_storage.snapshot_add([snapshot])

        with cook_extract_snapshot(swh_storage, snapshot.swhid()) as (ert, p):
            tag = ert.repo[b"refs/tags/v1.1.0"]
            assert tag.as_raw_string() == malformed_rel_manifest

            commit = ert.repo[tag.object[1]]
            assert commit.as_raw_string() == malformed_rev_manifest

            tree = ert.repo[commit.tree]
            assert tree.as_raw_string() == malformed_dir_manifest



# Copyright (C) 2020-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from functools import partial
import os
from typing import Any, Dict

import pkg_resources.extern.packaging.version
import pytest
from pytest_postgresql import factories

from swh.core.db.pytest_plugin import initialize_database_for_module
from swh.vault import get_vault
from swh.vault.backend import VaultBackend

os.environ["LC_ALL"] = "C.UTF-8"

# needed for directory tests on git-cloned repositories
# 022 is usually the default value, but some environments (eg. Debian builds) have
# a different one.
os.umask(0o022)

pytest_v = pkg_resources.get_distribution("pytest").parsed_version
if pytest_v < pkg_resources.extern.packaging.version.parse("3.9"):

    @pytest.fixture
    def tmp_path():
        import pathlib
        import tempfile

        with tempfile.TemporaryDirectory() as tmpdir:
            yield pathlib.Path(tmpdir)


vault_postgresql_proc = factories.postgresql_proc(
    load=[
        partial(initialize_database_for_module, "vault", VaultBackend.current_version)
    ],
)

postgres_vault = factories.postgresql("vault_postgresql_proc")


@pytest.fixture
def swh_vault_config(postgres_vault, tmp_path) -> Dict[str, Any]:
    tmp_path = str(tmp_path)
    return {
        "db": postgres_vault.dsn,
        "storage": {
            "cls": "memory",
        },
        "cache": {
            "cls": "pathslicing",
            "root": tmp_path,
            "slicing": "0:1/1:5",
            "allow_delete": True,
        },
        "scheduler": {
            "cls": "remote",
            "url": "http://swh-scheduler:5008",
        },
    }


@pytest.fixture
def swh_vault(swh_vault_config):
    return get_vault("local", **swh_vault_config)


@pytest.fixture
def swh_storage(swh_vault):
    return swh_vault.storage



# Copyright (C) 2017-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from typing import Dict

import pytest
import yaml

from swh.vault.cookers import COOKER_TYPES, get_cooker
from swh.vault.tests.test_backend import TEST_SWHID


@pytest.fixture
def swh_cooker_config():
    return {
        "vault": {
            "cls": "remote",
            "args": {
                "url": "mock://vault-backend",
                "storage": {"cls": "remote", "url": "mock://storage-url"},
            },
        }
    }


def write_config_to_env(config: Dict, tmp_path, monkeypatch) -> str:
    """Write the configuration dict into a temporary file, then reference that path to
    SWH_CONFIG_FILENAME environment variable.

    """
    conf_path = os.path.join(str(tmp_path), "cooker.yml")
    with open(conf_path, "w") as f:
        f.write(yaml.dump(config))
    monkeypatch.setenv("SWH_CONFIG_FILENAME", conf_path)
    return conf_path


def test_write_to_env(swh_cooker_config, tmp_path, monkeypatch):
    actual_path = write_config_to_env(swh_cooker_config, tmp_path, monkeypatch)

    assert os.path.exists(actual_path) is True
    assert os.environ["SWH_CONFIG_FILENAME"] == actual_path

    with open(actual_path, "r") as f:
        actual_config = yaml.safe_load(f.read())
    assert actual_config == swh_cooker_config


@pytest.mark.parametrize(
    "config_ko,exception_class,exception_msg",
    [
        ({}, ValueError, "missing 'vault' configuration"),
        (
            {"vault": {"cls": "local"}},
            EnvironmentError,
            "This vault backend can only be a 'remote' configuration",
        ),
        (
            {"vault": {"cls": "remote", "args": {"missing-storage-key": ""}}},
            ValueError,
            "invalid configuration: missing 'storage' config entry",
        ),
    ],
)
def test_get_cooker_config_ko(
    config_ko, exception_class, exception_msg, monkeypatch, tmp_path
):
    """Misconfigured cooker should fail the instantiation with exception message"""
    write_config_to_env(config_ko, tmp_path, monkeypatch)

    with pytest.raises(exception_class, match=exception_msg):
        get_cooker("flat", TEST_SWHID)


@pytest.mark.parametrize(
    "config_ok",
    [
        {
            "vault": {
                "cls": "remote",
                "args": {
                    "url": "mock://vault-backend",
                    "storage": {"cls": "remote", "url": "mock://storage-url"},
                },
            }
        },
        {
            "vault": {
                "cls": "remote",
                "args": {
                    "url": "mock://vault-backend",
                },
            },
            "storage": {"cls": "remote", "url": "mock://storage-url"},
        },
        {
            "vault": {
                "cls": "remote",
                "url": "mock://vault-backend",
            },
            "storage": {"cls": "remote", "url": "mock://storage-url"},
        },
    ],
)
def test_get_cooker_nominal(config_ok, tmp_path, monkeypatch):
    """Correct configuration should allow the instantiation of the cookers"""
    for cooker_type in COOKER_TYPES.keys():
        write_config_to_env(config_ok, tmp_path, monkeypatch)

        cooker = get_cooker(cooker_type, TEST_SWHID)

        assert cooker is not None
        assert isinstance(cooker, tuple(COOKER_TYPES[cooker_type]))



# Copyright (C) 2017-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

import os
from typing import Any, Dict, List, Type

from swh.core.config import load_named_config
from swh.core.config import read as read_config
from swh.model.swhids import CoreSWHID, ObjectType
from swh.storage import get_storage
from swh.vault import get_vault
from swh.vault.cookers.base import DEFAULT_CONFIG, DEFAULT_CONFIG_PATH, BaseVaultCooker
from swh.vault.cookers.directory import DirectoryCooker
from swh.vault.cookers.git_bare import GitBareCooker
from swh.vault.cookers.revision_flat import RevisionFlatCooker
from swh.vault.cookers.revision_gitfast import RevisionGitfastCooker

_COOKER_CLS: List[Type[BaseVaultCooker]] = [
    DirectoryCooker,
    RevisionFlatCooker,
    RevisionGitfastCooker,
    GitBareCooker,
]
COOKER_TYPES: Dict[str, List[Type[BaseVaultCooker]]] = {}


for _cooker_cls in _COOKER_CLS:
    COOKER_TYPES.setdefault(_cooker_cls.BUNDLE_TYPE, []).append(_cooker_cls)


def get_cooker_cls(bundle_type: str, object_type: ObjectType):
    cookers = COOKER_TYPES.get(bundle_type)

    if not cookers:
        raise ValueError(f"{bundle_type} is not a valid bundle type.")

    for cooker in cookers:
        try:
            cooker.check_object_type(object_type)
        except ValueError:
            pass
        else:
            return cooker

    raise ValueError(
        f"{object_type.name.lower()} objects do not have a {bundle_type} cooker"
    )


def check_config(cfg: Dict[str, Any]) -> Dict[str, Any]:
    """Ensure the configuration is ok to run a vault worker, and propagate defaults

    Raises:
        EnvironmentError if the configuration is not for remote instance
        ValueError if one of the following keys is missing: vault, storage

    Returns:
        New configuration dict to instantiate a vault worker instance

    """
    cfg = cfg.copy()

    if "vault" not in cfg:
        raise ValueError("missing 'vault' configuration")

    vcfg = cfg["vault"]
    if vcfg["cls"] != "remote":
        raise EnvironmentError(
            "This vault backend can only be a 'remote' configuration"
        )

    # TODO: Soft-deprecation of args key. Remove when ready.
    vcfg.update(vcfg.get("args", {}))

    # Default to top-level value if any
    if "storage" not in vcfg:
        vcfg["storage"] = cfg.get("storage")

    if not vcfg.get("storage"):
        raise ValueError("invalid configuration: missing 'storage' config entry.")

    return cfg


def get_cooker(bundle_type: str, swhid: CoreSWHID):
    """Instantiate a cooker class of type bundle_type.

    Returns:
        Cooker class in charge of cooking the bundle_type with id swhid.

    Raises:
        ValueError in case of a missing top-level vault key configuration or a storage
          key.
        EnvironmentError in case the vault configuration reference a non remote class.

    """
    if "SWH_CONFIG_FILENAME" in os.environ:
        cfg = read_config(os.environ["SWH_CONFIG_FILENAME"], DEFAULT_CONFIG)
    else:
        cfg = load_named_config(DEFAULT_CONFIG_PATH, DEFAULT_CONFIG)
    cooker_cls = get_cooker_cls(bundle_type, swhid.object_type)

    cfg = check_config(cfg)
    vcfg = cfg["vault"]

    storage = get_storage(**vcfg.pop("storage"))
    backend = get_vault(**vcfg)

    try:
        from swh.graph.client import RemoteGraphClient  # optional dependency

        graph = RemoteGraphClient(**vcfg["graph"]) if vcfg.get("graph") else None
    except ModuleNotFoundError:
        if vcfg.get("graph"):
            raise EnvironmentError(
                "Graph configuration required but module is not installed."
            )
        else:
            graph = None

    kwargs = {
        k: v for (k, v) in cfg.items() if k in ("max_bundle_size", "thread_pool_size")
    }

    return cooker_cls(
        swhid,
        backend=backend,
        storage=storage,
        graph=graph,
        **kwargs,
    )



# Copyright (C) 2016-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

import os
from typing import Any, Dict, Optional

from swh.core.api import RPCServerApp
from swh.core.api import encode_data_server as encode_data
from swh.core.api import error_handler
from swh.core.config import config_basepath, merge_configs, read_raw_config
from swh.vault import get_vault as get_swhvault
from swh.vault.backend import NotFoundExc
from swh.vault.interface import VaultInterface

from .serializers import DECODERS, ENCODERS

# do not define default services here
DEFAULT_CONFIG = {
    "client_max_size": 1024**3,
}


def get_vault():
    global vault
    if not vault:
        vault = get_swhvault(**app.config["vault"])

    return vault


class VaultServerApp(RPCServerApp):
    extra_type_decoders = DECODERS
    extra_type_encoders = ENCODERS


vault = None
app = VaultServerApp(
    __name__,
    backend_class=VaultInterface,
    backend_factory=get_vault,
)


@app.errorhandler(NotFoundExc)
def argument_error_handler(exception):
    return error_handler(exception, encode_data, status_code=400)


@app.errorhandler(Exception)
def my_error_handler(exception):
    return error_handler(exception, encode_data)


@app.route("/")
def index():
    return "SWH Vault API server"


def check_config(cfg: Dict[str, Any]) -> Dict[str, Any]:
    """Ensure the configuration is ok to run a postgresql vault server, and propagate
    defaults.

    Raises:
        EnvironmentError if the configuration is not for postgresql instance
        ValueError if one of the following keys is missing: vault, cache, storage,
        scheduler

    Returns:
        New configuration dict to instantiate a postgresql vault server instance.

    """
    cfg = cfg.copy()

    if "vault" not in cfg:
        raise ValueError("missing 'vault' configuration")

    vcfg = cfg["vault"]
    if vcfg["cls"] not in ("local", "postgresql"):
        raise EnvironmentError(
            "The vault backend can only be started with a 'postgresql' configuration",
        )

    # TODO: Soft-deprecation of args key. Remove when ready.
    vcfg.update(vcfg.get("args", {}))

    # Default to top-level value if any
    vcfg = {**cfg, **vcfg}

    for key in ("cache", "storage", "scheduler"):
        if not vcfg.get(key):
            raise ValueError(f"invalid configuration: missing {key} config entry.")

    return vcfg


def make_app_from_configfile(
    config_path: Optional[str] = None, **kwargs
) -> VaultServerApp:
    """Load and check configuration if ok, then instantiate (once) a vault server
    application.

    """
    config_path = os.environ.get("SWH_CONFIG_FILENAME", config_path)
    if not config_path:
        raise ValueError("Missing configuration path.")
    if not os.path.isfile(config_path):
        raise ValueError(f"Configuration path {config_path} should exist.")

    app_config = read_raw_config(config_basepath(config_path))
    app_config["vault"] = check_config(app_config)
    app.config.update(merge_configs(DEFAULT_CONFIG, app_config))

    return app


if __name__ == "__main__":
    print("Deprecated. Use swh-vault ")



# Copyright (C) 2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

pytest_plugins = [
    "swh.storage.pytest_plugin",
    "swh.loader.pytest_plugin",
]



# Copyright (C) 2020-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.model.from_disk import DentryPerms
from swh.model.model import Content, Directory, DirectoryEntry, SkippedContent
from swh.vault.to_disk import DirectoryBuilder, get_filtered_files_content


def test_get_filtered_files_content(swh_storage):
    content = Content.from_data(b"foo bar")
    skipped_content = SkippedContent(
        sha1=None,
        sha1_git=b"c" * 20,
        sha256=None,
        blake2s256=None,
        length=42,
        status="absent",
        reason="for some reason",
    )
    swh_storage.content_add([content])
    swh_storage.skipped_content_add([skipped_content])

    files_data = [
        {
            "status": "visible",
            "sha1": content.sha1,
            "sha1_git": content.sha1_git,
            "target": content.sha1_git,
        },
        {
            "status": "absent",
            "target": skipped_content.sha1_git,
        },
    ]

    res = list(get_filtered_files_content(swh_storage, files_data))

    assert res == [
        {
            "content": content.data,
            "status": "visible",
            "sha1": content.sha1,
            "sha1_git": content.sha1_git,
            "target": content.sha1_git,
        },
        {
            "content": (
                b"This content has not been retrieved in the "
                b"Software Heritage archive due to its size."
            ),
            "status": "absent",
            "target": skipped_content.sha1_git,
        },
    ]


def test_get_filtered_files_content__unknown_status(swh_storage):
    content = Content.from_data(b"foo bar")
    swh_storage.content_add([content])

    files_data = [
        {
            "status": "visible",
            "sha1": content.sha1,
            "sha1_git": content.sha1_git,
            "target": content.sha1_git,
        },
        {
            "status": "blah",
            "target": b"c" * 20,
        },
    ]

    with pytest.raises(AssertionError, match="unexpected status 'blah'"):
        list(get_filtered_files_content(swh_storage, files_data))


def _fill_storage(swh_storage, exclude_cnt3=False, exclude_dir1=False):
    cnt1 = Content.from_data(b"foo bar")
    cnt2 = Content.from_data(b"bar baz")
    cnt3 = Content.from_data(b"baz qux")
    dir1 = Directory(
        entries=(
            DirectoryEntry(
                name=b"content1",
                type="file",
                target=cnt1.sha1_git,
                perms=DentryPerms.content,
            ),
            DirectoryEntry(
                name=b"content2",
                type="file",
                target=cnt2.sha1_git,
                perms=DentryPerms.content,
            ),
        )
    )
    dir2 = Directory(
        entries=(
            DirectoryEntry(
                name=b"content3",
                type="file",
                target=cnt3.sha1_git,
                perms=DentryPerms.content,
            ),
            DirectoryEntry(
                name=b"subdirectory",
                type="dir",
                target=dir1.id,
                perms=DentryPerms.directory,
            ),
        )
    )
    if exclude_cnt3:
        swh_storage.content_add([cnt1, cnt2])
    else:
        swh_storage.content_add([cnt1, cnt2, cnt3])
    if exclude_dir1:
        swh_storage.directory_add([dir2])
    else:
        swh_storage.directory_add([dir1, dir2])

    return dir2


def test_directory_builder(swh_storage, tmp_path):
    dir2 = _fill_storage(swh_storage)

    root = tmp_path / "root"
    builder = DirectoryBuilder(swh_storage, bytes(root), dir2.id)

    assert not root.exists()

    builder.build()

    assert root.is_dir()
    assert set(root.glob("**/*")) == {
        root / "subdirectory",
        root / "subdirectory" / "content1",
        root / "subdirectory" / "content2",
        root / "content3",
    }

    assert (root / "subdirectory" / "content1").open().read() == "foo bar"
    assert (root / "subdirectory" / "content2").open().read() == "bar baz"
    assert (root / "content3").open().read() == "baz qux"


def test_directory_builder_missing_content(swh_storage, tmp_path):
    dir2 = _fill_storage(swh_storage, exclude_cnt3=True)

    root = tmp_path / "root"
    builder = DirectoryBuilder(swh_storage, bytes(root), dir2.id)

    assert not root.exists()

    builder.build()

    assert root.is_dir()

    assert "This content is missing" in (root / "content3").open().read()


def test_directory_builder_missing_directory(swh_storage, tmp_path):
    dir2 = _fill_storage(swh_storage, exclude_dir1=True)

    root = tmp_path / "root"
    builder = DirectoryBuilder(swh_storage, bytes(root), dir2.id)

    assert not root.exists()

    builder.build()

    assert root.is_dir()
    assert set(root.glob("**/*")) == {
        root / "subdirectory",
        root / "content3",
    }

    assert (root / "content3").open().read() == "baz qux"



# Copyright (C) 2016-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections
import functools
import os
from typing import Any, Dict, Iterator, List

from swh.model import hashutil
from swh.model.from_disk import DentryPerms, mode_to_perms
from swh.storage.algos.dir_iterators import dir_iterator
from swh.storage.interface import StorageInterface

MISSING_MESSAGE = (
    b"This content is missing from the Software Heritage archive "
    b"(or from the mirror used while retrieving it)."
)

SKIPPED_MESSAGE = (
    b"This content has not been retrieved in the "
    b"Software Heritage archive due to its size."
)

HIDDEN_MESSAGE = b"This content is hidden."


def get_filtered_files_content(
    storage: StorageInterface, files_data: List[Dict]
) -> Iterator[Dict[str, Any]]:
    """Retrieve the files specified by files_data and apply filters for skipped
    and missing contents.

    Args:
        storage: the storage from which to retrieve the objects
        files_data: list of file entries as returned by directory_ls()

    Yields:
        The entries given in files_data with a new 'content' key that points to
        the file content in bytes.

        The contents can be replaced by a specific message to indicate that
        they could not be retrieved (either due to privacy policy or because
        their sizes were too big for us to archive it).

    """
    for file_data in files_data:
        status = file_data["status"]
        if status == "visible":
            sha1 = file_data["sha1"]
            data = storage.content_get_data(sha1)
            if data is None:
                content = SKIPPED_MESSAGE
            else:
                content = data
        elif status == "absent":
            content = SKIPPED_MESSAGE
        elif status == "hidden":
            content = HIDDEN_MESSAGE
        elif status is None:
            content = MISSING_MESSAGE
        else:
            assert False, (
                f"unexpected status {status!r} "
                f"for content {hashutil.hash_to_hex(file_data['target'])}"
            )

        yield {"content": content, **file_data}


def apply_chunked(func, input_list, chunk_size):
    """Apply func on input_list divided in chunks of size chunk_size"""
    for i in range(0, len(input_list), chunk_size):
        yield from func(input_list[i : i + chunk_size])


class DirectoryBuilder:
    """Reconstructs the on-disk representation of a directory in the storage."""

    def __init__(self, storage: StorageInterface, root: bytes, dir_id: bytes):
        """Initialize the directory builder.

        Args:
            storage: the storage object
            root: the path where the directory should be reconstructed
            dir_id: the identifier of the directory in the storage
        """
        self.storage = storage
        self.root = root
        self.dir_id = dir_id

    def build(self) -> None:
        """Perform the reconstruction of the directory in the given root."""
        # Retrieve data from the database.
        # Split into files, revisions and directory data.
        entries = collections.defaultdict(list)
        for entry in dir_iterator(self.storage, self.dir_id):
            entries[entry["type"]].append(entry)

        # Recreate the directory's subtree and then the files into it.
        self._create_tree(entries["dir"])
        self._create_files(entries["file"])
        self._create_revisions(entries["rev"])

    def _create_tree(self, directories: List[Dict[str, Any]]) -> None:
        """Create a directory tree from the given paths

        The tree is created from `root` and each given directory in
        `directories` will be created.
        """
        # Directories are sorted by depth so they are created in the
        # right order
        bsep = os.path.sep.encode()
        directories = sorted(directories, key=lambda x: len(x["path"].split(bsep)))
        for dir in directories:
            os.makedirs(os.path.join(self.root, dir["path"]))

    def _create_files(self, files_data: List[Dict[str, Any]]) -> None:
        """Create the files in the tree and fetch their contents."""
        f = functools.partial(get_filtered_files_content, self.storage)
        files_data = apply_chunked(f, files_data, 1000)

        for file_data in files_data:
            path = os.path.join(self.root, file_data["path"])
            self._create_file(path, file_data["content"], file_data["perms"])

    def _create_revisions(self, revs_data: List[Dict[str, Any]]) -> None:
        """Create the revisions in the tree as broken symlinks to the target
        identifier."""
        for file_data in revs_data:
            path = os.path.join(self.root, file_data["path"])
            target = hashutil.hash_to_bytehex(file_data["target"])
            self._create_file(path, target, mode=DentryPerms.symlink)

    def _create_file(
        self, path: bytes, content: bytes, mode: int = DentryPerms.content
    ) -> None:
        """Create the given file and fill it with content."""
        perms = mode_to_perms(mode)
        if perms == DentryPerms.symlink:
            os.symlink(content, path)
        else:
            with open(path, "wb") as f:
                f.write(content)
            os.chmod(path, perms.value)



# Copyright (C) 2017-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict, List, Optional, Tuple

from typing_extensions import Protocol, runtime_checkable

from swh.core.api import remote_api_endpoint
from swh.model.swhids import CoreSWHID


@runtime_checkable
class VaultInterface(Protocol):
    """
    Backend Interface for the Software Heritage vault.
    """

    @remote_api_endpoint("fetch")
    def fetch(self, bundle_type: str, swhid: CoreSWHID) -> Optional[bytes]:
        """Fetch information from a bundle"""
        ...

    @remote_api_endpoint("cook")
    def cook(
        self, bundle_type: str, swhid: CoreSWHID, email: Optional[str] = None
    ) -> Dict[str, Any]:
        """Main entry point for cooking requests. This starts a cooking task if
        needed, and add the given e-mail to the notify list"""
        ...

    @remote_api_endpoint("progress")
    def progress(self, bundle_type: str, swhid: CoreSWHID):
        ...

    # Cookers endpoints

    @remote_api_endpoint("set_progress")
    def set_progress(self, bundle_type: str, swhid: CoreSWHID, progress: str) -> None:
        """Set the cooking progress of a bundle"""
        ...

    @remote_api_endpoint("set_status")
    def set_status(self, bundle_type: str, swhid: CoreSWHID, status: str) -> bool:
        """Set the cooking status of a bundle"""
        ...

    @remote_api_endpoint("put_bundle")
    def put_bundle(self, bundle_type: str, swhid: CoreSWHID, bundle):
        """Store bundle in vault cache"""
        ...

    @remote_api_endpoint("send_notif")
    def send_notif(self, bundle_type: str, swhid: CoreSWHID):
        """Send all the e-mails in the notification list of a bundle"""
        ...

    # Batch endpoints

    @remote_api_endpoint("batch_cook")
    def batch_cook(self, batch: List[Tuple[str, str]]) -> int:
        """Cook a batch of bundles and returns the cooking id."""
        ...

    @remote_api_endpoint("batch_progress")
    def batch_progress(self, batch_id: int) -> Dict[str, Any]:
        """Fetch information from a batch of bundles"""
        ...



# Copyright (C) 2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.vault import get_vault
from swh.vault.api.client import RemoteVaultClient
from swh.vault.backend import VaultBackend

SERVER_IMPLEMENTATIONS = [
    ("remote", RemoteVaultClient, {"url": "localhost"}),
    (
        "local",
        VaultBackend,
        {
            "db": "something",
            "cache": {"cls": "memory", "args": {}},
            "storage": {"cls": "remote", "url": "mock://storage-url"},
            "scheduler": {"cls": "remote", "url": "mock://scheduler-url"},
        },
    ),
]


@pytest.fixture
def mock_psycopg2(mocker):
    mocker.patch("swh.vault.backend.psycopg2.pool")
    mocker.patch("swh.vault.backend.psycopg2.extras")


def test_init_get_vault_failure():
    with pytest.raises(ValueError, match="Unknown Vault class"):
        get_vault("unknown-vault-storage")


@pytest.mark.parametrize("class_name,expected_class,kwargs", SERVER_IMPLEMENTATIONS)
def test_init_get_vault(class_name, expected_class, kwargs, mock_psycopg2):
    concrete_vault = get_vault(class_name, **kwargs)
    assert isinstance(concrete_vault, expected_class)


@pytest.mark.parametrize("class_name,expected_class,kwargs", SERVER_IMPLEMENTATIONS)
def test_init_get_vault_deprecation_warning(
    class_name, expected_class, kwargs, mock_psycopg2
):
    with pytest.warns(DeprecationWarning):
        concrete_vault = get_vault(class_name, args=kwargs)
    assert isinstance(concrete_vault, expected_class)


def test_init_get_vault_ok(swh_vault_config):
    concrete_vault = get_vault("local", **swh_vault_config)
    assert isinstance(concrete_vault, VaultBackend)
